// fd.patchVisualiser.js, by Frédéric Dufeu, 2013
// Javascript document to be used within fd.patchVisualiser.maxpat under Cycling'74 Max 6

mgraphics.init();
mgraphics.relative_coords = 0;
mgraphics.autofill = 0;

var patchRepresentationBuilt;
var abstractionName;

var jsuiWidth;
var jsuiHeight;
var displayOffsetX;
var displayOffsetY;
var displayZoomX;
var displayZoomY;

var thisPatcher;
var mainPatcher;

var r_mainPatcherProperties;
var r_layer;
var r_currentPatcher;
var r_currentPatcherProperties;

var r_cumulatedPatcherIndices;
var r_currentPatcherAddresses;

var r_parentPatcherIndices;
var r_currentPatcherIndices;

var numPatchers;
var numPatchersPerLayer;
var numObjects;

var cords;
var sizes;
var audiosizes;
var addresses;

var objectSizeX;
var objectSpacingX;
var objectSpacingY;

init();


function init(){
	patchRepresentationBuilt = false;
	abstractionName = 'fd.patchVisualiser';
	jsuiWidth = box.rect[2] - box.rect[0];
	jsuiHeight = box.rect[3] - box.rect[1];
	define_thisPatcher();
	define_mainPatcher();
	objectSizeX = 3;
	objectSpacingX = 2;
	objectSpacingY = 80;
	bang();
}
init.local = 1;


function buildPatchRepresentation(){
	mainRecursiveParsing();
}


function mainRecursiveParsing(){
	numPatchers = 0;
	numPatchersPerLayer = [];
	numObjects = 0;
	cords = [];
	sizes = [];
	audiosizes = [];
	addresses = [];
	r_currentPatcherAddresses = [];
	r_cumulatedPatcherIndices = [];
	r_parentPatcherIndices = [];
	r_currentPatcherIndices = [];
	r_layer = 0;
	if ( mainPatcher.name != abstractionName ){
		var maxnumpatchers = 0;
		numPatchers++;
		r_mainPatcherProperties = returnPatcherProperties(mainPatcher);
		r_currentPatcher = mainPatcher;
		r_currentPatcherProperties = r_mainPatcherProperties;
		numPatchers += r_currentPatcherProperties.numpatchers;
		numPatchersPerLayer[r_layer] = 1;
		numObjects += r_currentPatcherProperties.numobjects;
		sizes[r_layer] = [r_currentPatcherProperties.numobjects];
		audiosizes[r_layer] = [r_currentPatcherProperties.numaudioobjects];
		addresses[r_layer] = [[0]];
		r_currentPatcherAddresses[r_layer] = 0;
		r_currentPatcherIndices[r_layer] = 0;
		r_cumulatedPatcherIndices[r_layer] = 0;
		r_recursiveParsing();
		post('Total Number of Patchers', numPatchers, '\n');
		post('Total Number of Objects', numObjects, '\n');
		patchRepresentationBuilt = true;
		displayOffsetX = jsuiWidth * 0.5;
		displayOffsetY = 80;
		for ( var index = 0; index < numPatchersPerLayer.length; index++ ){
			maxnumpatchers = Math.max(maxnumpatchers, numPatchersPerLayer[index]);
		}
		displayZoomX = jsuiWidth / ( maxnumpatchers * ( objectSizeX + objectSpacingX ) * 1.05);
		displayZoomY = jsuiHeight / ( ( numPatchersPerLayer.length - 1 ) * objectSpacingY * 1.05 );
		outlet(0, 'displayOffsetX', displayOffsetX);
		outlet(0, 'displayOffsetY', displayOffsetY);
		outlet(0, 'displayZoomX', displayZoomX);
		outlet(0, 'displayZoomY', displayZoomY);
		bang();
	}
}
mainRecursiveParsing.local = 1;


function r_recursiveParsing(){
	var subpatcheraddresses = r_currentPatcherProperties.patcheraddresses;
	r_layer++;
	if ( numPatchersPerLayer[r_layer] == null ){
		numPatchersPerLayer[r_layer] = subpatcheraddresses.length;
	} else {
		numPatchersPerLayer[r_layer] += subpatcheraddresses.length;
	}
	if ( r_parentPatcherIndices[r_layer - 1] == null ){
		r_parentPatcherIndices[r_layer - 1] = 0;
	} else {
		r_parentPatcherIndices[r_layer - 1]++;
	}
	for ( var index = 0; index < subpatcheraddresses.length; index++ ){
		var address = subpatcheraddresses[index];
		if ( r_currentPatcherIndices[r_layer] == null ){
			r_currentPatcherIndices[r_layer] = 0;
		} else {
			r_currentPatcherIndices[r_layer]++;
		}
		r_currentPatcherAddresses[r_layer] = address;
		r_currentPatcher = returnMaxobjAtPatcherAddress(r_currentPatcher, address).subpatcher();
		if ( r_currentPatcher.name == abstractionName ){
			numPatchersPerLayer[r_layer]--;
			numPatchers--;
			numObjects--;
		} else {
			var cord = {};
			var patcherdeepaddress = [0];
			r_currentPatcherProperties = r_mainPatcherProperties;
			for ( var index2 = 0; index2 < r_layer - 1; index2++ ){ // DIRT: INDEX2
				r_currentPatcherProperties = r_currentPatcherProperties.patchers[0];
			}
			r_currentPatcherProperties.patchers.push(returnPatcherProperties(r_currentPatcher));
			r_currentPatcherProperties = r_currentPatcherProperties.patchers[r_currentPatcherProperties.patchers.length - 1];
			numPatchers += r_currentPatcherProperties.numpatchers;
			numObjects += r_currentPatcherProperties.numobjects;
			cord.origin_layer = r_layer - 1;
			cord.origin_index = r_parentPatcherIndices[r_layer - 1];
			cord.destination_layer = r_layer;
			cord.destination_index = r_currentPatcherIndices[r_layer];
			cords.push(cord);
			if ( sizes[r_layer] == null ){
				sizes[r_layer] = [];
				audiosizes[r_layer] = [];
				addresses[r_layer] = [];
			}
			sizes[r_layer].push(r_currentPatcherProperties.numobjects);
			audiosizes[r_layer].push(r_currentPatcherProperties.numaudioobjects);
			addresses[r_layer].push([].concat(r_currentPatcherAddresses));
			r_recursiveParsing();
		}
	}
	r_layer--;
	r_currentPatcher = r_currentPatcher.parentpatcher;
}
r_recursiveParsing.local = 1;


function paint(){
	with ( mgraphics ){
		set_source_rgba(1, 1, 1, 1);
		paint();
		if ( patchRepresentationBuilt ){
			identity_matrix();
			translate(displayOffsetX, displayOffsetY);
			scale(displayZoomX, displayZoomY);
			set_source_rgba(0, 0, 0, 1);
			set_line_width(0.125);
			for ( index = 0; index < cords.length; index++ ){
				var cord, origin_x, origin_y, destination_x, destination_y;
				cord = cords[index];
				origin_x = ( cord.origin_index * ( objectSizeX + objectSpacingX )) - (( numPatchersPerLayer[cord.origin_layer] * objectSizeX + ( numPatchersPerLayer[cord.origin_layer] - 1 ) * objectSpacingX ) * 0.5 ) + ( objectSizeX * 0.5 );
				origin_y = cord.origin_layer * objectSpacingY;
				destination_x = ( cord.destination_index * ( objectSizeX + objectSpacingX )) - (( numPatchersPerLayer[cord.destination_layer] * objectSizeX + ( numPatchersPerLayer[cord.destination_layer] - 1 ) * objectSpacingX ) * 0.5 ) + ( objectSizeX * 0.5 );
				destination_y = cord.destination_layer * objectSpacingY;
				move_to(origin_x, origin_y);
				line_to(destination_x, destination_y);
				stroke();
			}
			set_line_width(0.5);
			for ( index = 0; index < numPatchersPerLayer.length; index++ ){
				var numpatchers;
				numpatchers = numPatchersPerLayer[index];
				for ( index2 = 0; index2 < numpatchers; index2++ ){
					var patchersize, audiosize, left_x, right_x, top_y, bottom_y;
					patchersize = sizes[index][index2];
					audiosize = audiosizes[index][index2];
					left_x = ( index2 * ( objectSizeX + objectSpacingX )) - (( numpatchers * objectSizeX + ( numpatchers - 1 ) * objectSpacingX ) * 0.5 );
					right_x = left_x + objectSizeX;
					top_y = index * objectSpacingY - ( patchersize * 0.5 );
					if ( patchersize ){
						bottom_y = top_y + patchersize + 2;
					} else {
						bottom_y = top_y;
					}
					move_to(left_x, top_y);
					line_to(right_x, top_y);
					line_to(right_x, bottom_y);
					line_to(left_x, bottom_y);
					close_path();
					if ( audiosize ){
						set_source_rgba(1, 0, 0, 0.5);
						fill_preserve();
						set_source_rgba(1, 0, 0, 1);
						stroke();
					} else {
						set_source_rgba(0, 0, 0, 1);
						stroke();
					}
				}
			}	
		}
	}
}
paint.local = 1;


function bang(){
	mgraphics.redraw();
}
bang.local = 1;


function move_display(x, y){
	displayOffsetX = x;
	displayOffsetY = y;
	bang();
}


function zoom_display(x, y){
	displayZoomX = x;
	displayZoomY = y;
	bang();
}


function onclick(x, y){
	if ( patchRepresentationBuilt ){
		var abs_x, abs_y, layer, layer_float, patcherindex, patcherindex_float, patcherdeepaddresses, patcher, inrange;
		abs_x = ( x - displayOffsetX ) / displayZoomX;
		abs_y = ( y - displayOffsetY ) / displayZoomY;
		inrange = false;
		layer_float = ( abs_y + ( objectSpacingY * 0.5 )) / objectSpacingY;
		layer = Math.floor(layer_float);
		if (( layer > -1 ) && ( layer < numPatchersPerLayer.length - 1 )){
			patcherindex_float = ( abs_x + (( numPatchersPerLayer[layer] * objectSizeX + ( numPatchersPerLayer[layer] - 1 ) * objectSpacingX ) * 0.5 )) / ( objectSizeX + objectSpacingX );
			patcherindex = Math.floor(patcherindex_float);
			patcherdeepaddresses = addresses[layer][patcherindex];
			if (( patcherindex > -1 ) && ( patcherindex < numPatchersPerLayer[layer] )){
				if ( sizes[layer][patcherindex] ){
					if (( layer_float - layer > ( 0.5 - ( sizes[layer][patcherindex] * 0.5 / objectSpacingY ))) && ( layer_float - layer < ( 0.5 + ( sizes[layer][patcherindex] * 0.5 / objectSpacingY )))){
						inrange = true;
					}
				} else {
					if (( layer_float - layer > 0.4 ) && ( layer_float - layer < 0.6 )){
						inrange = true;
					}
				}
			}
			if ( inrange ){
				for ( var index = 0; index < layer + 1; index++ ){
					var patcheraddress = patcherdeepaddresses[index];
					if ( index == 0 ){
						patcher = mainPatcher;
					} else {
						patcher = returnMaxobjAtPatcherAddress(patcher, patcheraddress).subpatcher();
					}
				}
				patcher.front();	
			}
		}
	}
}
onclick.local = 1;


function define_thisPatcher(){
	thisPatcher = this.patcher;
}
define_thisPatcher.local = 1;


function define_mainPatcher(){
	mainPatcher = thisPatcher;
	while ( mainPatcher.box ){
		mainPatcher = mainPatcher.parentpatcher;
	}
}
define_mainPatcher.local = 1;


function returnPatcherProperties(patcher){
	var patcherproperties = {};
	var maxobjcounter = 0;
	var maxobj = patcher.firstobject;
	var numobjects = patcher.count;
	var numpatchers = 0;
	var patcheraddresses = [];
	var numaudioobjects = 0;
	while ( maxobj ){
		if ( maxobj.maxclass == 'patcher' ){
			numpatchers++;
			patcheraddresses.push(maxobjcounter);
			if ( maxobj.subpatcher().name.charAt(maxobj.subpatcher().name.length - 1) == '~' ){
				numaudioobjects++;
			}
		} else if ( maxobj.maxclass.charAt(maxobj.maxclass.length - 1) == '~' ){
			numaudioobjects++;
		}
		maxobj = maxobj.nextobject;
		maxobjcounter++;
	}
	patcherproperties.numobjects = numobjects;
	patcherproperties.numpatchers = numpatchers;
	patcherproperties.patcheraddresses = patcheraddresses;
	patcherproperties.patchers = [];
	patcherproperties.numaudioobjects = numaudioobjects;
	return patcherproperties;
}
returnPatcherProperties.local = 1;


function returnMaxobjAtPatcherAddress(patcher, address){
	var maxobj = patcher.firstobject;
	for ( var index = 0; index < address; index++ ){
		maxobj = maxobj.nextobject;
	}
	return maxobj;
}
returnMaxobjAtPatcherAddress.local = 1;